// fmcperf.cxx	-- compute fuel, distance, eta
//
// Written by Harald Johnsen, started Aug 2004.
//
// Copyright (C) 2004  Harald Johnsen  - hjohnsen@evc.net
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//


#include <simgear/compiler.h>
#include <simgear/math/sg_geodesy.hxx>
#include <simgear/constants.h>


#ifdef SG_HAVE_STD_INCLUDES
#  include <cstdio>
#  include <cstdlib>
#  include <ctime>
#else
#  include <math.h>
#  include <stdio.h>
#  include <stdlib.h>
#endif

#include "fmc.hxx"
#include "fmcdata.hxx"
#include "fmcperf.hxx"

#include <math.h>

/** Knots to Metter per second */
#define SG_KT_TO_METERPS        0.51444563

static const float g = 9.81;				// gravitational constant



/*



fmc/autopilot/active-spd

fmc/autopilot/active-spd-mode

	LRC SPD - Long range cruise speed
	ECON SPD - Economy speed
	SEL SPD - Selected speed manually entered on the CDU
	LIM SPD - Speed is limited by VMO, MMO, flap limit, or buffet limit
	MCP SPD - MCP speed entered on the MCP IAS/MACH indicator
	EO SPD - Engine out speed
	CO SPD - Engine out operations at airline specified engine out company speed
	VREF +80, for engine out operations during takeoff.

*/
static const string STR="";


void FMCPerf::update(void)
{
	// fake values for debug
	DataSet("fmc/autopilot/active-spd-mode", "SEL");
	DataSet("fmc/autopilot/active-spd", "200");
}

void FMCPerf::init(void)
{
	string ac = DataGet("ac-model");

	fdmS = 1135;
	fdmCLalpha = 4.4;
	fdmCL0 = 0.2;
	fdmCLmax = 1.2;
	fdmCD0 = 0.02;
	fdmK = 0.043;
	fdmTSFC = 0.657;

	if(ac == "747-100")
	{
		fdmS	= 5500;
		fdmTSFC = 0.657;	// JT9D-3
		DataSet("fmc/perf/zf-wt", "370");
	}
	else if(ac == "737")
	{
		fdmS	= 1171;
		fdmTSFC = 0.657;	// CFM56
		fdmCL0 = 0.2;
		fdmCD0 = 0.02;
		fdmK = 0.043;
		DataSet("fmc/perf/zf-wt", "83");
	}
	else if(ac == "747")
	{
		fdmS	= 5000;
		fdmTSFC = 0.576;	// CFM56
		DataSet("fmc/perf/zf-wt", "370");
	}
	else if(ac == "777-200LR")
	{
		fdmS	= 4605;
		fdmTSFC = 0.57;	// GE90-110B1
		fdmCL0 = 0.2;
		fdmCD0 = 0.02;
		fdmK = 0.043;
		DataSet("fmc/perf/zf-wt", "451.2");
	}
	else if(ac == "A320")
	{
		fdmS	= 1317;
		fdmTSFC = 0.657;	// CFM56_5
		DataSet("fmc/perf/zf-wt", "111");
	}
	else if(ac == "A340-500")
	{
		fdmS	= 4729;
		fdmTSFC = 0.8;	// RR-Trent-553
		fdmCL0 = 0.2;
		fdmCD0 = 0.017;
		fdmK = 0.043;
		DataSet("fmc/perf/zf-wt", "491.8");
	}
	else if(ac == "c172r")
	{
		fdmS	= 174;
		// dummy tsfc because its not a jet engine
		fdmTSFC = 1;	// engIO360C
		DataSet("fmc/perf/zf-wt", "1.5");
	}
}

FMCPerf::FMCPerf(void)
{
	// dummy init values
	fdmS = 1135;			//    wing area
	fdmCLalpha = 4.4;		//     'per radian
	fdmCL0 = 0.2;			//
	fdmCLmax = 1.2;
	fdmCD0 = 0.022;
	fdmK = 0.043;
	fdmTSFC = 0.657;
	fuel_eta_calc[1].zulu = "----";
	fuel_eta_calc[1].dist_to_go = 0;
	init();
}




float FMCPerf::calcCL(float Lift , float rho , float v , float S)
{
    return Lift / (0.5 * rho * v * v * S);
}
float FMCPerf::CalcCd(float cdo, float K, float cl, float M)
{
    // drag coefficient
    float Cd = cdo + K * cl * cl;
    // corrected for compressible flow effects
    return Cd / sqrtf(1.0f - M * M);
}


// h in meter
float FMCPerf::alt_dens(float h)
{
	float sigma;
    if(h < 11000.0)
        sigma = pow(1.0 - h * 0.000022558, 4.2559);
    else
        sigma = 0.2971 * expf((11000.0 - h) / 6341.5987);

    return sigma * 1.225;
}

// h in meter
float FMCPerf::alt_temp(float h)	// in Rankin
{
	float saTheta;

    if(h < 11000.0)
        saTheta = 1.0 - h / 145442;
    else
        saTheta = 0.751865;
    
    return 518.67 * saTheta;
}

float FMCPerf::velocity_mach(float v, float h)
{
	float soundVal;
	float tempK;

    tempK = alt_temp(h) / 1.8;
    soundVal = sqrtf(1.4 * 287.1 * tempK);
    return v / soundVal;
}

float FMCPerf::calcMach(float spd, float temp)
{
    return spd / sqrtf(1.4 * 287.1 * temp);
}

// Minimum Drag
// Clmd = sqr(CD0 * 1 / K )
// Vmd  = sqr(W / (0.5 * rho(h) * Clmd * S) )
// LtoDmax = sqr( 1 / ( K * 4 * CD0 ) )

// Vstall(h) = sqr( W / (0.5 * rho(h) * CLmax * S) )

float FMCPerf::computeVfromCl(float cl, float w, float h)
{
	cl += fdmCL0;
	return  sqrtf( (w * g / 2.2) / 
		(0.5 * alt_dens(h * SG_FEET_TO_METER) * cl * 
			fdmS * SG_FEET_TO_METER * SG_FEET_TO_METER) );
}

string FMCPerf::calcTime( int ct, long int warp ) 
{

#if defined(_MSC_VER) || defined(__MINGW32__)
    struct tm m_gmt;    // copy of system gmtime(&time_t) structure
    struct tm *gmt = &m_gmt;
#else
    struct tm *gmt;
#endif

    // get current Unix calendar time (in seconds)
    // warp += warp_delta;
    time_t cur_time;
    if ( ct ) {
        cur_time = ct + warp;
    } else {
        cur_time = time(NULL) + warp;
    }

    // get GMT break down for current time
#if defined(_MSC_VER) || defined(__MINGW32__)
    memcpy( gmt, gmtime(&cur_time), sizeof(tm) );
#else
    gmt = gmtime(&cur_time);
#endif

	char buf[20];
    sprintf( buf, "%02d%02d@sz@n", gmt->tm_hour, gmt->tm_min);
	return buf;
}

// v = knot, w = lb, alt = ft
float FMCPerf::compute_fuel(float v, float alt, float w)
{

	// Thrust Specific fuel compumption (TSFC)
	float tsfc = fdmTSFC / 60.0;
	//Altitude in meters
	alt *= SG_FEET_TO_METER;

	v *= SG_KT_TO_METERPS;	// knot to m/s
	//Mach number
	float M = velocity_mach(v, alt);
	if( M >= 1.0 )
		return 9e9;

	//air density calculated as a function of altitude
	float rho = alt_dens(alt);

	float S = fdmS * SG_FEET_TO_METER * SG_FEET_TO_METER;

	//lift coefficient required for L=W
	float cl = calcCL(w * g / 2.2, rho, v, S);

	//drag coefficient
	float cd = CalcCd(fdmCD0, fdmK, cl, M);

	//lift to drag ratio
	float lond = cl / cd;

	// thrust required
	float thrust = w / lond;
	// approximate fuel flow rate in lb/m
	float fuelflow = tsfc * thrust;

	return fuelflow;
}

// d = NM, v = knot, time = minutes
float FMCPerf::time_needed(float d, float v)
{
	d *= SG_NM_TO_METER;
	if ( v < 1.0 )
		v = 1.0;
	v *= SG_KT_TO_METERPS;
	return d / v / 60.0;
}

// ff = flow in lb/min, t = time in min, result = quantity in lb
float FMCPerf::fuel_needed(float ff, float t)
{
	return ff * t;
}

// no units
float FMCPerf::Cost_Function(float vk, float altft, float wlb)
{
	float ffmin = compute_fuel(vk, altft, wlb);
	float CI = 80;
	return ((CI * 100) + ffmin) / vk;
}

void FMCPerf::test_tuel(void)
{
	float res;

	res = Cost_Function(220, 15000, 100000);
	res = Cost_Function(300, 15000, 100000);

	res = Cost_Function(220, 25000, 100000);
	res = Cost_Function(300, 25000, 100000);

	res = Cost_Function(220, 35000, 100000);
	res = Cost_Function(300, 35000, 100000);

	res = Cost_Function(220, 38000, 100000);
	res = Cost_Function(300, 38000, 100000);
	res = Cost_Function(350, 38000, 100000);

}

float FMCPerf::query_wp_speed(string wp)
{
	string spd = DataGet(wp, false);
	if( spd.empty() || spd == "ECON" )
		spd = DataGet("fmc/perf/econ-spd", false);
	if( ! spd.empty() )
		return fabs( stof( spd ) );
	return 200.0;
}

float FMCPerf::query_wp_alt(string wp)
{
	string alt = DataGet(wp, false);
	if( alt.empty() )
		alt = DataGet("fmc/perf/crz-alt", false);
	if( ! alt.empty() )
		return fabs(stof( alt ) );
	return 10000.0;
}

void FMCPerf::fuel_eta_update(void)
{
	string pos;
	char prop[80];
	float reflat, reflon, lat, lon;
	double az1, az2, d;


	float fuel_remaining = stof( DataGet("fmc/perf/fuel-wt", false) ) * 1000.0;
	float zfw = stof( DataGet(("fmc/perf/zf-wt"), false) ) * 1000.0;	//83000.0; //370000+90000;
	float dist_to_go = 0;
	float total_time = 0.0;

	fuel_eta_calc[1].zulu = "----";
	fuel_eta_calc[1].dist_to_go = 0;

	pos = DataGet("fmc/fmc-pos", false);
	extract_latlon ( pos, &reflat, &reflon );

    int legnr = 1;
	do
	{
		sprintf(prop, "fmc/legmod/wp[%d]", legnr);
        string dest = DataGet( STR + prop + "/id", false);
		if( dest.empty() )
			break;

        pos = DataGet(STR +prop + "/pos", false);
        float this_wp_alt = query_wp_alt(STR +prop + "/alt" );
        float this_wp_spd = query_wp_speed(STR + prop + "/spd") ;

		extract_latlon( pos, &lat, &lon );

        geo_inverse_wgs_84(450, reflat, reflon, lat, lon, &az1, &az2, &d);
        int dist = d * SG_METER_TO_NM;
        int dir = az1;

		float weight = zfw + fuel_remaining;
		float time = time_needed(dist, this_wp_spd);
		float fuel_flow = compute_fuel(this_wp_spd, this_wp_alt, weight);
		float fuel_consumed = fuel_needed(fuel_flow, time);

		fuel_remaining -= fuel_consumed;
		dist_to_go += dist;
		total_time += time * 60;

		fuel_eta_calc[legnr].dist_to_wp = dist;
		fuel_eta_calc[legnr].dist_to_go = dist_to_go;
		fuel_eta_calc[legnr].alt = this_wp_alt;
		fuel_eta_calc[legnr].eta = time;
		fuel_eta_calc[legnr].zulu = calcTime(0, total_time);
		fuel_eta_calc[legnr].fuel_remaining = fuel_remaining / 1000.0;
		fuel_eta_calc[legnr].speed = this_wp_spd;
		fuel_eta_calc[legnr].weight = weight;

		if(zfw <= 0.0 || fuel_remaining <= 0)
		{
			fuel_eta_calc[legnr].eta = 0;
			fuel_eta_calc[legnr].zulu = "----";
		}
		// now we move to next wp to calculate hdg and dist
        reflat = lat; reflon = lon;
		legnr ++;
    } while(legnr < 200);

}
